"""Test the gator.api.client module."""
import json
from pathlib import Path
from types import SimpleNamespace

import pytest
from pytest_httpserver import HTTPServer
from gator.core.schemas.timetable import CourseSchema, OrganisationSchema

from gator.api.client import GatorClient

fixtures_dir = Path(__file__).parent / 'fixtures'


class TestGatorClient:
    """Test the :class:`gator.api.client.GatorClient` class."""
    @staticmethod
    @pytest.mark.usefixtures('httpserver')
    @pytest.fixture
    def gator_client(httpserver: HTTPServer) -> GatorClient:
        """Configures a GatorClient for testing."""
        return GatorClient(httpserver.url_for('/'))

    @pytest.mark.usefixtures('gator_client', 'httpserver')
    def test_get_courses_all(self, gator_client: GatorClient,
                             httpserver: HTTPServer) -> None:
        """Test the get_courses API method with no ids parameter."""
        courses_fixture = json.load(open(fixtures_dir / 'courses_all.json'))
        httpserver.expect_request('/courses', method='GET').respond_with_json(
            courses_fixture)

        expected = SimpleNamespace(
            courses=CourseSchema(many=True).load(courses_fixture['courses']),
            last_id=courses_fixture['last_id'],
        )

        assert gator_client.get_courses() == expected

    @pytest.mark.usefixtures('gator_client', 'httpserver')
    def test_get_courses_some(self, gator_client: GatorClient,
                              httpserver: HTTPServer) -> None:
        """Test the get_courses API method with ids parameter."""
        courses_fixture = json.load(open(fixtures_dir / 'courses_some.json'))
        course_ids = [c['id'] for c in courses_fixture['courses']]

        httpserver.expect_request(
            '/courses',
            query_string=dict(ids=','.join(course_ids)),
            method='GET'
        ).respond_with_json(courses_fixture)

        expected = SimpleNamespace(
            courses=CourseSchema(many=True).load(courses_fixture['courses']),
            last_id=courses_fixture['last_id'],
        )

        # Test comma-separated input
        assert gator_client.get_courses(ids=','.join(course_ids)) == expected
        # Test list input
        assert gator_client.get_courses(ids=course_ids) == expected

    @pytest.mark.usefixtures('gator_client', 'httpserver')
    def test_get_course(self, gator_client: GatorClient,
                        httpserver: HTTPServer) -> None:
        """Test the get_course API method."""
        course_fixture = json.load(open(fixtures_dir / 'course_ACT350.json'))
        course_id = course_fixture['id']

        httpserver.expect_request(f'/courses/{course_id}', method='GET')\
            .respond_with_json(course_fixture)

        expected = CourseSchema().load(course_fixture)

        assert gator_client.get_course('ACT350H1-F-20229') == expected

    @pytest.mark.usefixtures('gator_client', 'httpserver')
    def test_get_organisations(self, gator_client: GatorClient,
                               httpserver: HTTPServer) -> None:
        """Test the get_organisations API method."""
        organisations_fixture = json.load(open(
            fixtures_dir / 'organisations.json'))

        httpserver.expect_request('/organisations', method='GET').respond_with_json(
            organisations_fixture)

        expected = SimpleNamespace(
            organisations=OrganisationSchema(many=True).load(
                organisations_fixture['organisations']),
            last_id=organisations_fixture['last_id'],
        )

        assert gator_client.get_organisations() == expected

    @pytest.mark.usefixtures('gator_client', 'httpserver')
    def test_get_organisation(self, gator_client: GatorClient,
                              httpserver: HTTPServer) -> None:
        """Test the get_organisation API method."""
        organisation_fixture = json.load(open(
            fixtures_dir / 'organisation_MAT.json'))

        organisation_code = organisation_fixture['code']

        httpserver.expect_request(f'/organisations/{organisation_code}', method='GET')\
            .respond_with_json(organisation_fixture)

        expected = OrganisationSchema().load(organisation_fixture)

        assert gator_client.get_organisation('MAT') == expected



"""API helpers."""
import json
from typing import Any, Optional
from urllib.parse import ParseResult, parse_qsl, unquote, urlencode, urlparse


def add_url_params(url: str, params: Optional[dict] = None, **kwargs: Any) -> str:
    """Add GET params to provided URL being aware of existing.

    Args:
        url: target URL to add params to
        params: dict of key:value pairs to be added as params
        kwargs: key:value pairs to be added as params as keyword arguments

    Remarks:
        If both params and kwargs are provided, the two are merged with kwargs
        taking precedence.

    Returns:
        URL with added params.

    Examples:
        >>> new_params = {'answers': False, 'data': ['some','values']}
        >>> add_url_params('https://foo.com', new_params)
        'https://foo.com?answers=false&data=some&data=values'
        >>> add_url_params('https://bar.com/hello', name='world')
        'https://bar.com/hello?name=world'
    """
    # Merge params and kwargs into one dict, with kwargs taking precedence
    params = params or {}
    params.update(kwargs)

    # Unquoting URL first so we don't loose existing args
    url = unquote(url)
    # Extracting url info
    parsed_url = urlparse(url)
    # Extracting URL arguments from parsed URL
    get_args = parsed_url.query
    # Converting URL arguments to dict
    parsed_get_args = dict(parse_qsl(get_args))
    # Merging URL arguments dict with new params
    parsed_get_args.update(params)

    # Bool and Dict values should be converted to json-friendly values
    # you may throw this part away if you don't like it :)
    parsed_get_args.update(
        {k: json.dumps(v) for k, v in parsed_get_args.items()
         if isinstance(v, (bool, dict))}
    )

    # Converting URL argument to proper query string
    encoded_get_args = urlencode(parsed_get_args, doseq=True)
    # Creating new parsed result object based on provided with new
    # URL arguments. Same thing happens inside of urlparse.
    new_url = ParseResult(
        parsed_url.scheme, parsed_url.netloc, parsed_url.path,
        parsed_url.params, encoded_get_args, parsed_url.fragment
    ).geturl()

    return new_url



"""Web client for the Gator API."""
import json
from types import SimpleNamespace
from typing import Any, Optional, Union
from urllib.parse import quote_plus, urljoin

import certifi
import urllib3
from gator.api.client.helpers import add_url_params
from gator.core.models.timetable import Course, Organisation
from gator.core.schemas.timetable import CourseSchema, OrganisationSchema


class GatorClient:
    """Web client for the Gator API."""

    def __init__(self, base_url: str, encoding: str = 'utf8') -> None:
        """Initialize the client.

        Args:
            base_url: The base URL of the Gator API service.
            encoding: The encoding to use for the requests.
        """
        self._base_url = base_url
        self._encoding = encoding
        self._connection_pool = urllib3.PoolManager(
            cert_reqs='CERT_REQUIRED',
            ca_certs=certifi.where()
        )

    # Private methods
    def _compose_url(self, path: str, **params: Any) -> str:
        """Compose a URL for the given path.

        Args:
            path: The path to compose the URL for. This path is relative to the
                base URL.
            **params: The parameters to include in the URL.

        Returns:
            The composed URL.

        Examples:
            >>> client = GatorClient('http://localhost:5000')
            >>> client._compose_url('/api/v1/users', name='John', age=42)
            'http://localhost:5000/api/v1/users?name=John&age=42'
        """
        url = urljoin(self._base_url, path)
        return add_url_params(url, **params)

    def _request(self, method: str, path: str, **params: Any) \
            -> tuple[Any, int]:
        """Make a request to the Gator API.

        Args:
            method: The HTTP method to use for the request.
            path: The path to the resource to request. This path is relative to
                the base URL.
            params: The parameters to include in the request given as a
                dictionary of key-value pairs. If None, no parameters are
                included in the request.

        Returns:
            A tuple containing the JSON response and the HTTP status code.
        """
        url = self._compose_url(path, **params)
        response = self._connection_pool.urlopen(method.upper(), url)

        body = json.loads(response.data.decode(self._encoding))
        return body, response.status

    def _construct_params(self, **kwargs: Any) -> dict:
        """Construct a dictionary of parameters from keyword arguments.

        If a keyword argument is None, it is not included in the resultant
        parameter dictionary.

        Args:
            **kwargs: The keyword arguments to include in the parameters.

        Returns:
            A dictionary of parameters.
        """
        params = {}
        for key, value in kwargs.items():
            if value is not None:
                params[key] = value
        return params

    # API methods
    def get_courses(self, ids: Optional[Union[list[str], str]] = None,
                    page_size: Optional[int] = None,
                    last_id: Optional[str] = None) -> SimpleNamespace:
        """GET /courses.

        Args:
            ids: The ids of the courses to retrieve, either as a list of
                strings or a single string of comma-separated ids. If
                specified, returns only those courses with the given ids in the
                order specified. Ignore courses that could not be found.
                Otherwise, return all courses, in no particular order.
            page_size: The number of items to return per page.
            last_id: The id of the last item returned.

        Returns:
            A SimpleNamespace with two attributes:
                - courses: A list of Course objects.
                - last_id: The id of the last item returned.
        """
        if ids is not None and isinstance(ids, list):
            ids = ','.join(ids)

        params = self._construct_params(
            page_size=page_size,
            last_id=last_id,
            ids=ids
        )

        body, _ = self._request('GET', '/courses', **params)
        body['courses'] = CourseSchema(many=True).load(body['courses'])

        return SimpleNamespace(**body)

    def get_course(self, id: str) -> Course:
        """GET /courses/{id}.

        Args:
            id: The id (full code) of the course to retrieve.

        Returns:
            The Course object.
        """
        id_url_safe = quote_plus(id)
        body, _ = self._request('GET', f'/courses/{id_url_safe}')

        return CourseSchema().load(body)

    def get_organisations(self, page_size: Optional[int] = None,
                          last_id: Optional[str] = None) -> SimpleNamespace:
        """GET /organisations.

        Args:
            page_size: The number of items to return per page.
            last_id: The id of the last item returned.

        Returns:
            A SimpleNamespace with two attributes:
                - organisations: A list of Organisation objects.
                - last_id: The id of the last item returned.
        """
        params = self._construct_params(
            page_size=page_size,
            last_id=last_id
        )

        body, _ = self._request('GET', '/organisations', **params)
        body['organisations'] = OrganisationSchema(many=True).load(
            body['organisations']
        )

        return SimpleNamespace(**body)

    def get_organisation(self, code: str) -> Organisation:
        """GET /organisations/{code}.

        Args:
            code: The code of the organisation to retrieve.

        Returns:
            The Organisation object.
        """
        code_url_safe = quote_plus(code)
        body, _ = self._request('GET', f'/organisations/{code_url_safe}')

        return OrganisationSchema().load(body)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['gator', 'gator.api.client']

package_data = \
{'': ['*']}

install_requires = \
['certifi>=2022.6.15,<2023.0.0',
 'gator-core>=1.1.4,<2.0.0',
 'urllib3>=1.26.9,<2.0.0']

setup_kwargs = {
    'name': 'gator-client',
    'version': '0.3.0',
    'description': 'A web client for the Gator API',
    'long_description': '# Gator Client\nA web client for the [Gator](https://github.com/sqrl-planner/gator) API.\n\n## Package manager\ngator-client uses the [poetry](https://python-poetry.org/) package manager to manage its dependencies. To install the dependencies, run the following command:\n```\npoetry install\n```\nSee the [poetry](https://python-poetry.org/) documentation for more information and\ninstallation instructions.\n\n<!-- ... Insert content here ... -->\n\n## Tools\n\nThere are a number of tools available to help you with the development of sqrl. These tools ensure that your code is well-formed, follows the best practices, and\nis consistent with the rest of the project.\n\n#### Linting the codebase\n- For detecting code quality and style issues, run ``flake8``\n- For checking compliance with Python docstring conventions, run ``pydocstyle``\n\n**NOTE**: these tools will not fix any issues, but they can help you identify potential problems.\n\n#### Formatting the codebase\n- For automatically formatting the codebase, run ``autopep8 --in-place --recursive .``. For more information on this command, see the [autopep8](https://pypi.python.org/pypi/autopep8) documentation.\n- For automatically sorting imports, run ``isort .``\n\n#### Running tests\nFor running tests, run ``pytest``.\n',
    'author': 'Shon Verch',
    'author_email': 'verchshon@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
